#include <stdio.h>
#include <string.h>
#include <math.h>

#include <aversive/irq_lock.h>
#include <aversive/wait.h>
#include <aversive/pgmspace.h>

#include <uart.h>

#include "mpu6050.h"
#include "matrix.h"
#include "MadgwickAHRS.h"
#include "sd_log.h"
#include "imu.h"
#include "main.h"

/* structure storing the latest imu infos returned by the sensor */
static struct imu_info g_imu;

/* structure storing the latest quaternion */
static struct quaternion g_quat;

/* structure storing the latest euler position */
static struct euler g_euler;

/* periodical timer structure */
static struct callout imu_timer;

static void quaternion2euler(const struct quaternion *quat, struct euler *euler)
{
	float q0 = quat->q0;
	float q1 = quat->q1;
	float q2 = quat->q2;
	float q3 = quat->q3;

	euler->roll = atan2f(2.0f * (q0 * q1 + q2 * q3),
		q0*q0 - q1*q1 - q2*q2 + q3*q3);
	euler->pitch = -asinf(2.0f * (q1 * q3 - q0 * q2));
	euler->yaw = atan2f(2.0f * (q1 * q2 + q0 * q3),
		q0*q0 + q1*q1 - q2*q2 - q3*q3);
}

/* timer callback that polls IMU and does the calculation */
static void imu_cb(struct callout_mgr *cm, struct callout *tim, void *arg)
{
	struct imu_info imu;
	struct quaternion quat;
	struct euler euler;
	uint8_t irq_flags;

	(void)arg;

	/* copy previous quaternion locally */
	IRQ_LOCK(irq_flags);
	memcpy(&quat, &g_quat, sizeof(quat));
	IRQ_UNLOCK(irq_flags);

	mpu6050_read_all_axes(&g_imu);
	MadgwickAHRSupdate(&g_imu, &g_quat);
	quaternion2euler(&g_quat, &g_euler);

	/* update global variables */
	IRQ_LOCK(irq_flags);
	memcpy(&g_imu, &imu, sizeof(g_imu));
	IRQ_UNLOCK(irq_flags);
	IRQ_LOCK(irq_flags);
	memcpy(&g_quat, &quat, sizeof(g_quat));
	IRQ_UNLOCK(irq_flags);
	IRQ_LOCK(irq_flags);
	memcpy(&g_euler, &euler, sizeof(g_euler));
	IRQ_UNLOCK(irq_flags);

	/* reschedule event */
	callout_schedule(cm, tim, 2);
}

void imu_init(void)
{
	mpu6050_init();

	callout_init(&imu_timer, imu_cb, NULL, IMU_PRIO);
	callout_schedule(&imuboard.intr_cm, &imu_timer, 20); /* every 20ms */
}

void imu_get_info(struct imu_info *imu)
{
	memcpy(imu, &g_imu, sizeof(*imu));
}

void imu_get_pos_quat(struct quaternion *quat)
{
	memcpy(quat, &g_quat, sizeof(*quat));
}

void imu_get_pos_euler(struct euler *euler)
{
	memcpy(euler, &g_euler, sizeof(*euler));
}


int imu_log(void)
{
	char buf[128];
	int16_t len;
	uint32_t ms;
	uint8_t flags;
	struct imu_info imu;

	if (sd_log_enabled() == 0)
		return 0;

	IRQ_LOCK(flags);
	ms = global_ms;
	imu_get_info(&imu);
	IRQ_UNLOCK(flags);

	len = snprintf(buf, sizeof(buf),
		"%"PRIu32"\t"
		"gyro %+3.3f\t%+3.3f\t%+3.3f\t\t"
		"accel %+3.3f\t%+3.3f\t%+3.3f\t\t"
		"magnet %+3.3f\t%+3.3f\t%+3.3f\r\n",
		ms,
		imu.gx, imu.gy, imu.gz,
		imu.ax, imu.ay, imu.az,
		imu.mx, imu.my, imu.mz);
	if (sd_log_write(buf, len) != len) {
		printf_P(PSTR("error writing to file\n"));
		return -1;
	}

	return 0;
}
